from __future__ import annotations

import shlex
import subprocess as sp
import time
from typing import Callable
from dataclasses import dataclass, asdict, field as data_field
from pathlib import Path
from typing import Any

from .aliases import Aliases
from .command import Command
from .logger import log
from .pipe import Pipe
from ._py_version import PY
from ._types import RunPipeCmd


@dataclass
class RunConfig:
    # run
    input: str | bytes | None
    capture_output: bool
    timeout: int | None
    check: bool
    # Popen
    bufsize: int
    executable: str | None
    stdin: int | None
    stdout: int | None
    stderr: int | None
    preexec_fn: Callable | None
    close_fds: bool
    shell: bool  # to expand before runing (eg **/*.py, $HOME, ...)
    cwd: str | None  # use it to change directory
    env: dict | None
    universal_newlines: None
    startupinfo: None
    creationflags: int
    restore_signals: bool
    start_new_session: bool
    pass_fds: tuple
    encoding: None
    errors: None
    text: None

    @classmethod
    def default(cls) -> "RunConfig":
        # fmt: off
        return cls(
            # run
            input=None, capture_output=True, timeout=None, check=False,
            # Popen
            bufsize=-1, executable=None, stdin=None, stdout=None,
            stderr=None, preexec_fn=None, close_fds=True, shell=False,
            cwd=None, env=None, universal_newlines=None, startupinfo=None,
            creationflags=0, restore_signals=True, start_new_session=False,
            pass_fds=(), encoding=None, errors=None, text=None,
        )
        # fmt: on


def do_nothing(*args, **kwargs):
    pass


@dataclass
class Shell:
    _cwd: Path = Path(__name__).parent
    fake: bool = False
    config: RunConfig = RunConfig.default()
    historic: list[Command] = data_field(default_factory=list)
    silent_piping: bool = True
    aliases = Aliases.default()

    def __post_init__(self, *args, **kwargs):
        self.config.cwd = str(self._cwd)

    def cd(self, path: Path | str):
        if isinstance(path, str):
            path = Path(path)

        if path.is_absolute:
            path = path.resolve()
        else:
            path = self.cd.joinpath(path).resolve()

        if not path.exists():
            raise sp.CalledProcessError(1, f"cd {str(path)}")

        self._cwd = path
        self.config.cwd = str(path)

    def cwd(self) -> Path:
        return self._cwd

    @staticmethod
    def _get_logger(silent: bool):
        return log.debug if silent else log.info

    def run(
        self, cmd: str, *args, fake=False, callback=do_nothing, silent=False, **kwargs
    ) -> Command:
        log_silent = self._get_logger(silent)
        command = Command(cmd, self._cwd, time.time(), -1.0, -1, None)
        self.historic.append(command)
        log_silent(f"❯ {cmd}")

        if self.fake or fake:
            print(f"fake ❯ {cmd}")
            command.return_code = 0
            command.end = command.start
            command.fake = True
            return command

        if PY >= "3.9":
            config = asdict(self.config) | kwargs
        else:
            config = asdict(self.config)
            config.update(kwargs)
        cmd = self.aliases.process_aliases(cmd)

        cmd_args = shlex.split(cmd) + list(args)
        if config.get("shell", False):
            cmd_args = shlex.split(cmd) + list(args)

        result: sp.CalledProcessError | sp.CompletedProcess | None = None
        try:
            completed_process: sp.CompletedProcess = sp.run(cmd_args, **config)
            result = completed_process
        except sp.CalledProcessError as exc:
            result = exc
            log.error(f"Command failed: {result.stderr}")
            raise exc
        finally:
            if result is not None:
                command.return_code = result.returncode
                command.result = result
                command.end = time.time()

                log_silent(command.stdout)
                callback(command=command)
        return command

    def __call__(self, *args, **kwargs):
        return self.run(*args, **kwargs)

    def __or__(self, cmd: RunPipeCmd) -> "Pipe":
        result = self.run_pipe(cmd)
        if isinstance(result, Command):
            return Pipe(
                sh=self,
                stdout=result.stdout,
                stderr=result.stderr,
                return_code=result.return_code,
            )
        else:
            return Pipe(sh=self, stdout=result)

    def __lshift__(self, stdin: Any) -> "Pipe":
        return Pipe(sh=self, stdout=stdin)

    def run_pipe(self, cmd: RunPipeCmd, **kwargs) -> Any | Command:
        if isinstance(cmd, str):
            if isinstance(kwargs.get("input"), str):
                kwargs["input"] = kwargs["input"].encode("utf-8")
            return self.run(cmd, silent=self.silent_piping, check=False, **kwargs)
        elif callable(cmd):
            stdin = kwargs.get("input")
            return cmd() if stdin is None else cmd(stdin)
        else:
            raise NotImplementedError(f"Shell does not support type {type(cmd)}.")



from _typeshed import Incomplete

def cmp(a: int, b: int) -> int: ...

class PythonVersion:
    version: Incomplete
    def compare(self, expected_str: str) -> int: ...
    def __eq__(self, expected: object) -> bool: ...
    def __lt__(self, expected: str) -> bool: ...
    def __le__(self, expected: str) -> bool: ...
    def __gt__(self, expected: str) -> bool: ...
    def __ge__(self, expected: str) -> bool: ...

PY: Incomplete



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, TYPE_CHECKING


from .command import Command
from .file_operations import write_file
from .utils import PipeExtractor
from .xargs import XArgs, XArgsMeta
from ._types import PipeCmd, FileType

if TYPE_CHECKING:
    from .shell import Shell


@dataclass
class Pipe:
    sh: Shell
    stdout: Any
    stderr: Any | None = None
    return_code: int = 0

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(stdout={repr(self.stdout)})"

    def __or__(self, cmd: PipeCmd | PipeExtractor) -> "Pipe" | Any:
        stdin = self.stdout
        stdout: Any

        if isinstance(cmd, PipeExtractor):
            return cmd(self)
        elif isinstance(cmd, XArgsMeta):
            stdout = cmd(stdin=stdin)
        elif isinstance(stdin, XArgs):
            stdout = stdin.run(cmd, shell=self.sh)
        elif isinstance(cmd, XArgs):
            cmd = cmd(stdin=stdin)
            stdout = cmd.run(stdin, shell=self.sh)
        else:
            result = self.sh.run_pipe(cmd, input=stdin)
            if isinstance(result, Command):
                return Pipe(
                    sh=self.sh,
                    stdout=result.stdout,
                    stderr=result.stderr,
                    return_code=result.return_code,
                )
            else:
                stdout = result

        return Pipe(sh=self.sh, stdout=stdout)

    def __gt__(self, file: FileType) -> Any:
        write_file(self.stdout, file, "w", fake=self.sh.fake)
        return self.stdout

    def __rshift__(self, file: FileType) -> Any:
        write_file(self.stdout, file, "a+", fake=self.sh.fake)
        return self.stdout



from .pipe import Pipe as Pipe
from _typeshed import Incomplete
from typing import Any

class PipeExtractor:
    attr: str
    def __call__(self, pipe: Pipe) -> Any: ...

class Stdout(PipeExtractor):
    attr: str

class Stderr(PipeExtractor):
    attr: str

class Stdin(PipeExtractor):
    attr: str

stdout: Incomplete
stderr: Incomplete
stdin: Incomplete



import logging

from rich.logging import RichHandler


FORMAT = "%(message)s"
logging.basicConfig(
    level="NOTSET",
    format=FORMAT,
    datefmt="[%X]",
    handlers=[RichHandler(rich_tracebacks=True, tracebacks_show_locals=True)],
)
log = logging.getLogger("bulot")
log.setLevel(logging.INFO)


if __name__ == "__main__":
    log.debug("debug")
    log.info("info")
    log.warning("warning")
    log.error("error")
    log.critical("critical")



import subprocess as sp
from ._types import FileType as FileType
from .file_operations import write_file as write_file
from pathlib import Path
from typing import Any

class Command:
    value: str
    cwd: Path
    start: float
    end: float
    return_code: int
    result: Union[None, sp.CalledProcessError, sp.CompletedProcess]
    fake: bool
    @property
    def stdout(self) -> str: ...
    @property
    def stderr(self) -> str: ...
    def __gt__(self, file: FileType) -> Any: ...
    def __rshift__(self, file: FileType) -> Any: ...
    def __init__(self, value, cwd, start, end, return_code, result, fake) -> None: ...



from __future__ import annotations

import re
from collections import UserDict


ALIASES = {
    "ls": "ls --color=always",
    "grep": "grep --color=always --exclude-dir={.bzr,CVS,.git,.hg,.svn,.idea,.tox}",
}


class Aliases(UserDict):
    @classmethod
    def default(cls):
        return cls(ALIASES)

    @staticmethod
    def compile(alias):
        return re.compile(r"^\b" + alias + r"\b")

    def process_aliases(self, cmd: str):
        for (regex, replacement) in self.regexes.values():
            cmd = regex.sub(replacement, cmd)
        return cmd

    def __init__(self, *args, **kwargs):
        self.regexes = {}
        super().__init__(*args, **kwargs)

    def __setitem__(self, alias, replacement):
        self.data[alias] = replacement
        self.regexes[alias] = (Aliases.compile(alias), replacement)

    def __delitem__(self, alias):
        del self.data[alias]
        del self.regexes[alias]



from pathlib import Path


ECHO_STR = "Plop"
ECHO_CMD = f"echo '{ECHO_STR}'"
ECHO_ARGS = ["echo", ECHO_STR]
ECHO_STDOUT = f"{ECHO_STR}"

WRONG_FILE = "a_file_that_does_not_exist"
WRONG_CMD = f"ls {WRONG_FILE}"

PIPE_STDOUT = "Plop"
REPLACE_STDOUT = "truc"

FILE_NAME = "Plop.plop"
PATH_TESTS = Path(__file__).parent
PATH_DOES_NOT_EXIST = PATH_TESTS / "Plop"



from __future__ import annotations

from io import StringIO, BytesIO
from pathlib import Path
from typing import Callable, Union, TYPE_CHECKING


if TYPE_CHECKING:
    from .xargs import XArgs, XArgsMeta


class NotSet:
    pass


RunPipeCmd = Union[str, Callable]
PipeCmd = Union[str, Callable, "XArgs", "XArgsMeta"]
not_set = NotSet()

# if PY >= "3.10":
#     FileData = str | bytes
#     FileType = str | Path | StringIO | BytesIO
FileData = Union[str, bytes]
FileType = Union[str, Path, StringIO, BytesIO]
